/*
*    SPMD example using PVM 3.0
*/

#include <sys/types.h>
#include <sys/stat.h>
#include "pvm3.h"
#include <stdio.h>
#include <assert.h>

char *childArgs[] = {
    "-child",
    0,
};

int stay_busy();

int NPROC = 1;
int LOOPS = 500;
int parent = 0;

char msg[100];
#define BA_LENGTH 1024
double big_array[BA_LENGTH];
int group_instance, group_parent;
int atid, atag, alen;
int icount;

FILE *fd;

#ifdef USE_PVM_ERROR
             /* pvm_perror produces awkwardly formatted output from each node */
#define Xprintf pvm_perror
#else
             /* tag each message with a header identifying the sender node */
#define Xprintf if(parent){printf("MASTER: ");}else{printf("%06X: ",mytid);};\
         printf
#endif

main(argc, argv)
int argc;
char *argv[];
{
    int mytid;                  /* my task id */
    int tids[1000];            /* array of task id */
    int me;                     /* my process number */
    int i, sp;
    int in = 0;
    char *ptr;
    char msg_str[256];
    char my_name[80];
    char logdir[80];
    char *hostname;
    char *PDlib;

    strcpy(my_name,argv[0]);
    strncpy(logdir,"NULL",4);   /* initialize to NULL string */
    logdir[4]=0;

    /* enroll in pvm */
    mytid = pvm_mytid();

    /* kids use this */
    tids[0] = pvm_parent();

    sprintf(msg_str, "I am tid %06X running %s\n", mytid, my_name);
    Xprintf(msg_str);

    /* find out if I am parent or child */
    if (argc > 1)
    {
       /* I am the parent */
       Xprintf("I am the master!\n");
       parent = 1;
       Xprintf("argc > 1\n");
       tids[0] = mytid;
       me = 0;
       sprintf(msg_str,"args passed in = %d\n", argc);
       Xprintf(msg_str);
       if (argc >= 3)
	 {
             NPROC = (int) strtol (argv[1], &ptr, 10);
	     LOOPS =  (int) strtol (argv[2], &ptr, 10);
             sprintf(msg_str,"NPROC=%d, LOOPS=%d\n", NPROC, LOOPS);
             Xprintf(msg_str);
	 }
       if (argc>3)
         {
             strcpy(logdir,argv[3]);
             openlog(me,mytid,logdir);
         }

       /* start up copies of myself */
       if (NPROC < 2)
	  stay_busy();

       /* start the group */
       Xprintf("calling pvm_joingroup ...\n");
       assert (group_instance = pvm_joingroup ("potato_group") >= 0);
       sprintf(msg_str,"pvm_joingroup: group_instance=%d\n", group_instance);
       Xprintf(msg_str);

       sprintf(msg_str,"ready to spawn %d more copies of %s\n", NPROC, my_name);
       Xprintf(msg_str);
       sp = pvm_spawn(my_name, 0, 0, 0, NPROC, &tids[1]);
       for (i=1; i<=sp; ++i)
	{
          sprintf(msg_str, "Other tid is %06X\n", tids[i]);
	  Xprintf(msg_str);
	}

       if (sp != (NPROC))
	 {
	   Xprintf("spawn_failed\n");
	   pvm_exit();
	   exit(-1);
	 }
       else
	 {
           sprintf(msg_str,"SPAWNED %d:\n", sp);
           Xprintf(msg_str);
	 }

       /* multicast tids array to children */
       assert(pvm_initsend(PvmDataDefault)>= 0);
       assert(pvm_pkint(&NPROC, 1, 1) >= 0);
       assert(pvm_pkint(&LOOPS, 1, 1) >= 0);
       assert(pvm_pkint(tids, NPROC+1, 1) >= 0);
       assert(pvm_pkstr(logdir) >= 0);
       for (i=1; i<=NPROC; ++i)
       {
          assert(pvm_send(tids[i], 0) >= 0);
       }
       /* assert(pvm_mcast(&tids[1], NPROC, 0)>= 0); */
    }
    else    /* I am a child */
    {
       Xprintf("I am a child\n");
       /* receive tids array */
       if (pvm_recv(tids[0], -1) < 0)
	{
	pvm_exit();
	exit(0);
	}
       sprintf(msg_str,"My master is %06X\n",tids[0]);
       Xprintf(msg_str);
       Xprintf("msg received\n");
       if (pvm_upkint(&NPROC, 1, 1) < 0)
       {
	pvm_perror("NPROC not unpacked\n");
	pvm_exit();
	exit(0);
	}
       Xprintf("nproc unpacked\n");
       if (pvm_upkint(&LOOPS, 1, 1) < 0)
       {
	pvm_perror("LOOPS not unpacked\n");
	pvm_exit();
	exit(0);
	}
       Xprintf("loops unpacked\n");
       if (pvm_upkint(tids, NPROC+1, 1) < 0)
       {
	pvm_perror("TIDS not unpacked\n");
	pvm_exit();
	exit(0);
       }
       Xprintf("tids unpacked\n");
       for( i=1; i<=NPROC ; i++ )
          if( mytid == tids[i] ){ me = i; break; }

       if (pvm_upkstr(logdir) < 0)
       {
	pvm_perror("logdir not unpacked\n");
	pvm_exit();
	exit(0);
       } else {
        Xprintf("logdir unpacked\n");
        if (strncmp(logdir,"NULL",4)!=0) openlog(me,mytid,logdir);
       }

       assert (group_instance = pvm_joingroup ("potato_group") >= 0);
       Xprintf("ready to play!\n");
     }

    hostname=(char *)getenv("HOSTNAME");
    PDlib=(char *)getenv("PARADYN_LIB");
    sprintf(msg_str,"[%s]PARADYN_LIB=%s\n", hostname, PDlib);
    Xprintf(msg_str);

/*-----------------------------------------------------------
 * all NPROC tasks are equal now
 * and can address each other by tids[0] thru tids[NPROC]
 * for each process me => process number [0-(NPROC)]
 *-----------------------------------------------------------
 */

    /* initialize big_array using group functions */
    if (parent)
      {
	/* wait for everyone to join the group */
	while (pvm_gsize("potato_group") != NPROC+1) ;
	pvm_barrier("potato_group", NPROC+1);

	for (i=0; i<BA_LENGTH; ++i)
	  big_array[i] = drand48();

	assert(pvm_initsend(PvmDataDefault) >= 0);
	assert(pvm_pkdouble(big_array, BA_LENGTH, 1) >= 0);
	assert ((sp = pvm_bcast ("potato_group", 10)) >= 0);
	group_parent = group_instance;
      }
    else
      {
	pvm_barrier("potato_group", NPROC+1);
	assert( pvm_precv (-1, 10, (char *) big_array, BA_LENGTH,
		   PVM_DOUBLE, &atid, &atag, &alen) >= 0);
      }
     sprintf(msg_str,"Starting %d iterations\n",LOOPS);
     Xprintf(msg_str);
     icount = 0;
     for (i=0; i < LOOPS; i++)
       {
	 dowork( me, tids, NPROC+1 );
	 icount++;
	 if (icount == 100)
	   {
	     if (me == 0)
	       {
                 sprintf(msg_str,"Iteration %d of %d\n",i+1,LOOPS);
		 Xprintf(msg_str);
		 /*Xprintf("PARENT IN LOOP\n"); */
		 icount = 0;
	       }
	     /* do group operations */
	     /* do_group(me, tids, NPROC+1); */
	   }
       }
     Xprintf("token ring done\n");

     if (fd!=NULL) fclose(fd);

     /* program finished exit pvm */
     pvm_exit();
     exit(1);
}

/* use group operations from pvm */
do_group (me, tids, nproc)
{
  /* make sure that all processes are ready */
  pvm_barrier("potato_group", nproc);

  /*  */
  pvm_scatter(big_array, big_array, BA_LENGTH/nproc, PVM_DOUBLE,
	      12, "potato_group", group_parent);

  /* make sure that all processes are ready */
  pvm_barrier("potato_group", nproc);

  /* */
  pvm_gather(big_array, big_array, BA_LENGTH/nproc, PVM_DOUBLE,
	     13, "potato_group", group_parent); 

  /* make sure that all processes are ready */
  pvm_barrier("potato_group", nproc);

  pvm_reduce(PvmSum, big_array, BA_LENGTH, PVM_DOUBLE,
	     14, "potato_group", group_parent);

  /* make sure that all processes are ready */
  pvm_barrier("potato_group", nproc);
}

/* Simple example passes a token around a ring */
dowork( me, tids, nproc )
     int me;
     int *tids;
     int nproc;
{
     int i;
     int token;
     int dest;
     int count  = 1;
     int stride = 1;
     int msgtag = 4;
     static int call_count=0;


     if (fd!=NULL) { /* generate some fairly-pointless I/O */
        fprintf(fd,"dowork(me=%d,nproc=%d): call_count=%d\n",
                me,nproc,call_count);
        fflush(fd);
     }
     call_count++;
     if( me == 0 )
     { 
        token = tids[0];
        pvm_initsend( PvmDataDefault );
        pvm_pkint( &token, count, stride );
        pvm_send( tids[1], msgtag );
        pvm_recv( tids[nproc-1], msgtag );
     }
     else
     {
        pvm_recv( tids[me-1], msgtag );
        pvm_upkint( &token, count, stride );
        pvm_initsend( PvmDataDefault );
        pvm_pkint( &token, count, stride );
        dest = (me == nproc-1)? tids[0] : tids[me+1] ;
        pvm_send( dest, msgtag );
     }
     if (fd!=NULL) { /* generate some fairly-pointless I/O */
        fprintf(fd,"dowork(me=%d,nproc=%d): call_count=%d (done)\n",
                me,nproc,call_count);
        fflush(fd);
     }
}


int stay_busy()
{
int x, y;
float f;

for (x=0; x < LOOPS; ++x)
   {
      pvm_initsend(PvmDataDefault);
      pvm_pkint(&x, 1, 1);
      pvm_send(pvm_mytid(), x);
      pvm_recv(pvm_mytid(), x);
      pvm_upkint(&y, 1, 1);
      assert(x == y);
      printf("LOOP=%d\n", x);
   }
pvm_exit();
exit(0);

}

openlog (me, mytid, dirname)
int me, mytid;
char *dirname;
{
    struct stat statBuf;
    char logfile[80];
    char msg_str[128];
    sprintf(msg_str,"openlog(%s)\n",dirname);
    Xprintf(msg_str);
    if (stat(dirname, &statBuf) != 0) {     /* doesn't already exist */
        sprintf(msg_str,"Log directory does not exist: %s\n",dirname);
        Xprintf(msg_str);
        return;
    }
    if (!S_ISDIR(statBuf.st_mode)) {        /* not a directory! */
        sprintf(msg_str,"Log %s is not a directory!\n",dirname);
        Xprintf(msg_str);
        return;
    }
    sprintf(msg_str,"Log directory %s OK!\n",dirname);
    Xprintf(msg_str);
    sprintf(logfile,"%s/potato_%02d.log",dirname,me);
    fd=fopen(logfile,"w");
    if (fd==NULL) {
        sprintf(msg_str,"Warning: failed to open log file %s!\n", logfile);
        Xprintf(msg_str);
    } else {
        sprintf(msg_str,"Opened log file %s!\n", logfile);
        Xprintf(msg_str);
    }
}
