

#include <stdio.h>
#include <stdlib.h>
#include <thread.h>
#include <malloc.h>
#include <assert.h>
#include <sys/time.h>

#define DIM 500
#define MAX_PARALLEL 20
#define SCHED_FUDGE 1.2
#define TRUE 1
#define FALSE 0


#define MREF(mt, row, col) mt->m[(row) * mt->n_cols + (col)]

typedef struct {
  int n_rows ;
  int n_cols ;
  float m[1] ;
} matrix_t;

typedef struct {
  int i;
  int n;
  matrix_t *a, *b, *c;
  float trace_ans ;
} mm_params ;

int work (int x) {
  int i, s ;
  for (i=0; i<x; i++)
    s = s + i ;
  return s ;
}

float innerp (matrix_t *a, matrix_t* b, matrix_t *c, int row, int col) {
  int j ;
  float sum = 0.0 ;

  work(1000) ;

  for (j=0; j < a->n_cols; j++)
     sum += MREF(a, row, j) * MREF(b, j, col);
  MREF(c, row, col) = sum ;
 return sum ;
}

void* matmul_slave(void *param) {
  mm_params *p;
  matrix_t *a, *b, *c ;
  int row, col, j, k, quot, rem, do_rows, first_row;
  float trace ;
  int i ;
fprintf(stderr, "*** Thr_%d  Entering, matmul_slave...\n", thr_self());
  p = (mm_params *) param ;
  a = p->a ;
  b = p->b ;
  c = p->c ;
  trace = 0.0 ;

  quot = c->n_rows / p->n ;
  rem = c->n_rows % p->n ;
  do_rows = quot + ((p->i < rem)? 1 : 0);
  first_row = quot * p->i + ((p->i < rem)? p->i: rem) ;
  for (i=0; i< 20; i++) 
    for (row = first_row; row < first_row + do_rows ; row ++) {
      for (col = 0 ; col < c->n_cols; col++) {
        float sum =  innerp(a, b, c, row, col) ;
        if (row == col ) trace += sum ;
      }
    }
  p->trace_ans = trace ;
  return  (NULL) ;
}

#define min(a,b) (((a)>(b))?(b):(a))
float matrix_multiply_trace(matrix_t *a, matrix_t *b, matrix_t *c) {
  mm_params params[MAX_PARALLEL];
  thread_t thread[MAX_PARALLEL];
  int thread_present[MAX_PARALLEL];
  /* int n_threads = min (MAX_PARALLEL, (int) (sysconf(_SC_NPROCESSORS_ONLN) * SCHED_FUDGE)) ; */
  int n_threads = 4 ;
  int i, res;
  float trace = 0.0 ;

  thr_setconcurrency(n_threads+1) ;

  for (i=0; i < n_threads; i++) {
    thread_present[i] = TRUE ;
    params[i].i = i ;
    params[i].n = n_threads ;
    params[i].a = a ;
    params[i].b = b ;
    params[i].c = c ;

    switch (i) {
      case 0: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	      THR_BOUND, &thread[i]);
        fprintf(stderr, "after creating thr%d, THR_BOUND\n", thread[i]) ;
	break ;
      }
      case 1: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	      THR_DETACHED, &thread[i]);
        fprintf(stderr, "after creating thr%d, THR_DETACHED\n", thread[i]) ;
	break;
      }
      case 2: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	      THR_SUSPENDED, &thread[i]);
        fprintf(stderr, "after creating thr%d, THR_SUSPENDED\n", thread[i]) ;
	thr_continue(thread[i]) ;
	break;
      }
      case 3: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	      THR_NEW_LWP, &thread[i]);
        fprintf(stderr, "after creating thr%d, THR_NEW_LWP\n", thread[i]) ;
	break ;
      }
      case 4: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	      THR_DAEMON, &thread[i]);
        fprintf(stderr, "after creating thr%d, THR_DAEMON\n", thread[i]) ;
	break;
      }
      default: {
        res = thr_create(0, 0, matmul_slave, (void *) &params[i], 
	0, &thread[i]);
        fprintf(stderr, "after creating thr%d, 0\n", thread[i]) ;
        break ;
      }
    }
    if (res != 0) {
      thread_present[i] = FALSE ;
      matmul_slave(&params[i]);
    }
  }
  for (i=0; i< n_threads; i++) {
    if (thread_present[i])
      thr_join(thread[i], 0, 0) ;
    trace += params[i].trace_ans ;
  }
  return trace ;
}

matrix_t *a, *b, *c;

void init (int dim) {
  int i, j;
  a = malloc(sizeof (matrix_t) + sizeof(float)*(dim *dim -1)) ;
  b = malloc(sizeof (matrix_t) + sizeof(float)*(dim *dim -1)) ;
  c = malloc(sizeof (matrix_t) + sizeof(float)*(dim *dim -1)) ;
  assert(a);
  assert(b);
  assert(c);
  a->n_rows = b->n_rows = c->n_rows = dim ;
  a->n_cols = b->n_cols = c->n_cols = dim ;
  for (i=0; i< dim; i++) 
    for (j=0; j < dim; j++) {
      MREF(a, i, j) = (float) (i+j) ;
      MREF(b, i, j) = (float) (i+j) ;
      MREF(c, i, j) = (float) (i+j) ;
    }
      
}

int main (int argc, char *argv[]) {
  float trace ;
  hrtime_t start, end ;
  int k ;
  FILE* fp ;
  if (argc>1)
    fp = fopen(argv[1], "a+");
  else
    fp = stdout ;
  
  start = gethrtime() ;

  init (DIM) ;

     trace = matrix_multiply_trace(a, b, c) ;

  printf("trace = %f, Done\n", trace);
  fflush(stdout);
  end = gethrtime() ;
  fprintf(fp, "the elapsed time is %lld\n", end-start) ;
  fclose(fp);
}
